/* Copyright (C) 2024 FIAS Frankfurt Institute for Advanced Studies, Frankfurt / Main
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Felix Weiglhofer [committer], Dominik Smith */

#include "Unpack.h"

#include "log.hpp"

using namespace cbm::algo::trd2d;
using fles::Subsystem;

Unpack::Unpack(const ReadoutConfig& readout) : fReadout(readout)
{
  constexpr u8 SystemVersion = 0x02;

  // Create one algorithm per component for TRD and configure it with parameters
  auto equipIdsTrd2d = fReadout.GetEquipmentIds();
  for (auto& equip : equipIdsTrd2d) {

    trd2d::UnpackPar par{};
    const size_t numAsics = fReadout.GetNumAsics(equip);

    for (size_t asic = 0; asic < numAsics; asic++) {
      trd2d::UnpackAsicPar asicPar;
      const size_t numChans = fReadout.GetNumChans(equip, asic);

      for (size_t chan = 0; chan < numChans; chan++) {
        trd2d::UnpackChannelPar chanPar;
        auto pars           = fReadout.ChanMap(equip, asic, chan);
        chanPar.fPadAddress = pars.padAddress;    // Pad address for channel
        chanPar.fMask       = pars.rPairingFlag;  // Flag channel mask
        chanPar.fDaqOffset  = pars.daqOffset;     // Time calibration parameter
        asicPar.fChanParams.push_back(chanPar);
      }
      auto comppars         = fReadout.CompMap(equip);
      par.fSystemTimeOffset = fReadout.GetSystemTimeOffset();
      par.fModId            = comppars.moduleId;
      par.fCrobId           = comppars.crobId;
      par.fAsicParams.push_back(asicPar);
    }
    auto algo                      = std::make_unique<UnpackMS>(par);
    fAlgos[{equip, SystemVersion}] = std::move(algo);

    L_(debug) << "--- Configured equipment " << equip << " with " << numAsics << " asics";
  }
  L_(info) << "--- Configured " << fAlgos.size() << " unpacker algorithms for TRD2D.";
}

Unpack::Result_t Unpack::operator()(const fles::Timeslice& ts) const { return DoUnpack(Subsystem::TRD2D, ts); }
