/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfSetup.h
/// @brief  Setup representation for the Kalman-filter framework (header)
/// @since  28.03.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#pragma once  // include this header only once per compilation unit

#include "KfDefs.h"
#include "KfField.h"
#include "KfMaterialMap.h"
#include "KfTarget.h"
#include "KfVector.h"

#include <boost/serialization/access.hpp>
#include <boost/serialization/split_free.hpp>

#include <fstream>
#include <optional>
#include <sstream>
#include <string>
#include <vector>

namespace cbm::algo::kf
{
  /// \class  Setup
  /// \brief  KF-framework representation of the detector setup
  /// \tparam T  Underlying floating-point data type
  template<typename T>
  class alignas(VcMemAlign) Setup {
    friend class boost::serialization::access;  // Boost serializer methods
    template<typename>
    friend class Setup;


   public:
    /// \brief Constructor
    /// \param fldMode  Field mode
    Setup(EFieldMode fldMode)
      : fvMaterialLayers({})
      , fField(Field<T>(fldMode, EFieldType::Normal))
      , fTarget(Target<T>())
    {
    }

    /// \brief Destructor
    ~Setup() = default;

    /// \brief  Copy constructor
    /// \tparam I  Underlying floating-point data type of the source
    template<typename I>
    Setup(const Setup<I>& other)
      : fvMaterialLayers(other.fvMaterialLayers)
      , fField(other.fField)
      , fTarget(other.fTarget)
    {
    }

    /// \brief Move constructor
    //Setup(Setup&&) noexcept;

    /// \brief  Copy assignment operator
    Setup& operator=(const Setup& other) = default;

    /// \brief Move assignment operator
    //Setup& operator=(Setup&&) noexcept;

    /// \brief Sets material layer
    /// \param map     Material map
    void AddMaterial(const MaterialMap& map) { fvMaterialLayers.emplace_back(map); }

    /// \brief Makes an instance of the field depending on the template parameter
    /// \throw  std::logic_error  If the particular field member is undefined (nullopt)
    const Field<T>& GetField() const { return fField; }

    /// \brief Gets material layer
    /// \param iLayer  Index of layer
    const MaterialMap& GetMaterial(int iLayer) const { return fvMaterialLayers[iLayer]; }

    /// \brief Gets number of defined material layers
    int GetNofMaterialLayers() const { return static_cast<int>(fvMaterialLayers.size()); }

    /// \brief Gets target
    const Target<T>& GetTarget() const { return fTarget; }

    /// \brief String representation of the class contents
    /// \param verbosity    A verbose level for output
    /// \param indentLevel  Indent level of the string output
    std::string ToString(int verbosity = 1, int indentLevel = 0) const;

    /// \brief Sets field
    void SetField(const Field<double>& field) { fField = Field<T>(field); }

    /// \brief Sets target (moves from the source)
    /// \param target  Target instance
    void SetTarget(const Target<double>& target) { fTarget = target; }

   private:
    /// \brief BOOST serialization load method
    template<class Archive>
    void load(Archive& ar, const unsigned int /*version*/)
    {
      ar >> fvMaterialLayers;
      ar >> fTarget;
      ar >> fField;
    }

    template<class Archive>
    void save(Archive& ar, const unsigned int /*version*/) const
    {
      ar << fvMaterialLayers;
      ar << fTarget;
      ar << fField;
    }

    template<class Archive>
    void serialize(Archive& ar, const unsigned int version)
    {
      boost::serialization::split_member(ar, *this, version);
    }

    std::vector<MaterialMap> fvMaterialLayers = {};  ///< Container of the material maps
    Field<T> fField;                                 ///< Interpolated field (NOTE: maybe make optional)
    Target<T> fTarget;                               ///< Target layer
  };

}  // namespace cbm::algo::kf
