/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   KfSetupInitializer.h
/// @brief  A base KF-Setup initialization class (source)
/// @since  28.03.2024
/// @author Sergei Zharko <s.zharko@gsi.de>

#include "KfSetupInitializer.h"

using cbm::algo::kf::SetupInitializer;


// ---------------------------------------------------------------------------------------------------------------------
//
void SetupInitializer::AddMaterial(const MaterialMap& material)
{
  if (!fMaterialLayers.emplace(material).second) {
    std::stringstream msg;
    msg << "SetupInitializer::AddMaterial: attempt of adding a duplicating material layer " << material.ToString()
        << ".\nThe next material layers were already added:";
    for (const auto& el : fMaterialLayers) {
      msg << "\n\t- " << el.ToString();
    }
    throw std::logic_error(msg.str());
  }
}

// ---------------------------------------------------------------------------------------------------------------------
//
void SetupInitializer::Reset()
{
  fFieldFactory.Reset();
  fMaterialLayers.clear();
}

// ---------------------------------------------------------------------------------------------------------------------
//
void SetupInitializer::SetTargetProperties(double x, double y, double z, double fieldInitStep,
                                           const MaterialMap& material)
{
  fTarget.SetX(x);
  fTarget.SetY(y);
  fTarget.SetZ(z);
  fTarget.SetMaterial(material);
  fFieldFactory.SetTarget(x, y, z);
  fFieldFactory.SetStep(fieldInitStep);
}

// ---------------------------------------------------------------------------------------------------------------------
//
void SetupInitializer::Store(const Setup<double>& setup, const std::string& fileName)
{
  std::ofstream ofs(fileName, std::ios::binary);
  if (!ofs) {
    std::stringstream msg;
    msg << "kf::SetupInitializer::Store: failed openning file \"" << fileName << "\" to store the setup";
    throw std::runtime_error(msg.str());
  }
  boost::archive::binary_oarchive oa(ofs);
  oa << setup;
}
