/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmKfTrackingSetupInitializer.cxx
/// \brief  Tracking setup initializer in CBM (source)
/// \since  28.08.2024
/// \author Sergei Zharko <s.zharko@gsi.de>

#include "CbmKfTrackingSetupInitializer.h"

#include "CbmKfOriginalField.h"
#include "CbmKfTarget.h"
#include "CbmMuchTrackingInterface.h"
#include "CbmMvdTrackingInterface.h"
#include "CbmSetup.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTofTrackingInterface.h"
#include "CbmTrdTrackingInterface.h"
#include "FairField.h"
#include "FairRunAna.h"
#include "KfDefs.h"
#include "KfMaterialMapCreator.h"
#include "Logger.h"

#include <functional>

using cbm::kf::TrackingSetupInitializer;
using kf::tools::MaterialMapCreator;

// ---------------------------------------------------------------------------------------------------------------------
//
bool TrackingSetupInitializer::InitProperties()
try {
  using cbm::algo::kf::EFieldMode;
  using cbm::algo::kf::EFieldType;

  // Check, if a subsystem exists in the setup
  fbUseMvd &= CbmSetup::Instance()->IsActive(ECbmModuleId::kMvd);
  fbUseSts &= CbmSetup::Instance()->IsActive(ECbmModuleId::kSts);
  fbUseMuch &= CbmSetup::Instance()->IsActive(ECbmModuleId::kMuch);
  fbUseTrd &= CbmSetup::Instance()->IsActive(ECbmModuleId::kTrd);
  fbUseTof &= CbmSetup::Instance()->IsActive(ECbmModuleId::kTof);

  // Magnetic field initialization
  if (FairRunAna::Instance()->GetField()) {
    this->SetFieldProperty(EFieldMode::Intrpl, OriginalField(), EFieldType::Normal);
  }
  else {
    this->SetFieldProperty(EFieldMode::Intrpl, ZeroField(), EFieldType::Null);
  }

  // Tracking station property initialization
  {
    // Geometrical layer (contains information on the station sizes)
    //
    struct GeoLayer {
      double zRef;  // Reference z-coordinate [cm]
      double zMin;  // Lower z-coordinate boundary [cm]
      double zMax;  // Upper z-coordinate boundary [cm]
      double xMax;  // Half size of the layer in the x-direction [cm]
      double yMax;  // Half size of the layer in the y-direction [cm]
      bool operator<(const GeoLayer& rhs) const { return zRef < rhs.zRef; }
    };
    std::set<GeoLayer> geoStations;

    auto CollectStations = [&](const auto* pIfs) -> void {
      if (!pIfs) {
        return;
      }
      for (int iSt = 0; iSt < pIfs->GetNtrackingStations(); ++iSt) {
        geoStations.emplace(GeoLayer{.zRef = pIfs->GetZref(iSt),
                                     .zMin = pIfs->GetZmin(iSt),
                                     .zMax = pIfs->GetZmax(iSt),
                                     .xMax = std::max(std::fabs(pIfs->GetXmin(iSt)), std::fabs(pIfs->GetXmax(iSt))),
                                     .yMax = std::max(std::fabs(pIfs->GetYmin(iSt)), std::fabs(pIfs->GetYmax(iSt)))});
      }
    };
    CollectStations(fbUseMvd ? CbmMvdTrackingInterface::Instance() : nullptr);
    CollectStations(fbUseSts ? CbmStsTrackingInterface::Instance() : nullptr);
    CollectStations(fbUseMuch ? CbmMuchTrackingInterface::Instance() : nullptr);
    CollectStations(fbUseTrd ? CbmTrdTrackingInterface::Instance() : nullptr);
    CollectStations(fbUseTof ? CbmTofTrackingInterface::Instance() : nullptr);

    // Retriev target properties
    const auto* pTarget = Target::Instance();

    // Init material map creator
    MaterialMapCreator materialCreator{};
    materialCreator.SetSafeMaterialInitialization(kMatCreatorSafeMode);
    materialCreator.SetDoRadialProjection(pTarget->GetZ());
    materialCreator.SetNraysPerDim(kMatCreatorNrays);

    // Target initialization
    double zLast =
      pTarget->GetZ() + pTarget->GetDz() + kTargMaterialOffset;  // z-coordinate of the target material upper limit
    {
      // Material of the target
      double xyMax{1.3 * pTarget->GetRmax()};
      int nBins{std::min(static_cast<int>(std::ceil(2. * xyMax / kMatCreatorPitch)), kMatCreatorMaxNbins)};
      assert(nBins > 0);
      double zFirst{pTarget->GetZ() - pTarget->GetDz()};  // z-coordinate of the target material lower limit
      LOG(info) << "- collecting target material (z = " << zFirst << " - " << zLast << ", size = " << xyMax << ")\n";
      this->SetTargetProperties(pTarget->GetX(), pTarget->GetY(), pTarget->GetZ(), kTargFieldInitStep,
                                materialCreator.GenerateMaterialMap(pTarget->GetZ(), zFirst, zLast, xyMax, nBins));
    }

    // Material and field layers initialization

    for (const auto& layer : geoStations) {
      LOG(info) << " >>>>> LAYER >>>>> " << layer.zRef;
    }

    for (auto layerIt = geoStations.begin(); layerIt != geoStations.end(); ++layerIt) {
      double z1 = layerIt->zMax;
      double z2 = z1;
      if (std::next(layerIt) != geoStations.end()) {
        z2 = std::next(layerIt)->zMin;
      }
      double zNew  = 0.5 * (z1 + z2);
      double xyMax = 1.3 * std::max(std::fabs(layerIt->xMax), std::fabs(layerIt->yMax));
      int nBins    = std::min(static_cast<int>(std::ceil(2. * xyMax / kMatCreatorPitch)), kMatCreatorMaxNbins);
      assert(nBins > 0);

      LOG(info) << "- collecting material between z = " << zLast << " and " << zNew << " [cm]";
      this->AddMaterial(materialCreator.GenerateMaterialMap(layerIt->zRef, zLast, zNew, xyMax, nBins));
      this->AddFieldSliceReference(layerIt->xMax, layerIt->yMax, layerIt->zRef);
      zLast = zNew;
    }

    if constexpr (0) {
      // Estimate acceptance
      std::vector<cbm::algo::kf::MaterialMap> mapsTest;
      mapsTest.reserve(geoStations.size());
      double slope{0.};
      zLast =
        pTarget->GetZ() + pTarget->GetDz() + kTargMaterialOffset;  // z-coordinate of the target material upper limit
      for (auto layerIt = geoStations.begin(); layerIt != geoStations.end(); ++layerIt) {
        double xyMax = 1.3 * std::max(std::fabs(layerIt->xMax), std::fabs(layerIt->yMax));
        slope        = std::max(slope, xyMax / (layerIt->zRef - pTarget->GetZ()));
      }
      for (auto layerIt = geoStations.begin(); layerIt != geoStations.end(); ++layerIt) {
        double z1 = layerIt->zMax;
        double z2 = z1;
        if (std::next(layerIt) != geoStations.end()) {
          z2 = std::next(layerIt)->zMin;
        }
        double zNew  = 0.5 * (z1 + z2);
        double xyMax = slope * (layerIt->zRef - pTarget->GetZ());
        //int nBins    = std::min(static_cast<int>(std::ceil(2. * xyMax / kMatCreatorPitch)), kMatCreatorMaxNbins);
        int nBins = 100;
        assert(nBins > 0);

        LOG(info) << "* collecting material between z = " << zLast << " and " << zNew << " [cm]";
        mapsTest.push_back(materialCreator.GenerateMaterialMap(layerIt->zRef, zLast, zNew, xyMax, nBins));
        zLast = zNew;
      }
      for (const auto& map : mapsTest) {
        LOG(info) << "TEST: " << map.ToString(0, 2);
      }
      {
        // Addition test
        int iInactive{2};
        auto& M1{mapsTest[iInactive]};
        auto& M2{mapsTest[iInactive + 1]};
        auto M12gen{
          materialCreator.GenerateMaterialMap(M2.GetZref(), M1.GetZmin(), M2.GetZmax(), M2.GetXYmax(), M2.GetNbins())};
        auto M12add{M2};
        M12add.Add(M1);
        LOG(info) << "MATERIAL MAP ADDITION TEST:";
        LOG(info) << "M1=\n" << M1.ToString(0, 2);
        LOG(info) << "M2=\n" << M2.ToString(0, 2);
        LOG(info) << "M12gen=\n" << M2.ToString(0, 2);
        LOG(info) << "M12add=\n" << M2.ToString(0, 2);
      }
    }
  }
  LOG(info) << "kf::TrackingSetupInitializer: Tracking setup was initialized successfully";
  return true;
}
catch (const std::exception& err) {
  LOG(error) << "kf::TrackingSetupInitializer: Tracking setup was not initialized. Reason: " << err.what();
  return false;
}

// ---------------------------------------------------------------------------------------------------------------------
//
void TrackingSetupInitializer::Use(bool mvd, bool sts, bool much, bool trd, bool tof)
{
  fbUseMvd  = mvd;
  fbUseSts  = sts;
  fbUseMuch = much;
  fbUseTrd  = trd;
  fbUseTof  = tof;
}
