/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmKfTrackingSetupInitializer.h
/// \brief  Tracking setup initializer in CBM (source)
/// \since  28.08.2024
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "KfSetupInitializer.h"

#include <tuple>

namespace cbm::kf
{
  /// \class TrackingSetupInitializer
  /// \brief Encapsulation of the kf::Setup initialization routines for CBM
  class TrackingSetupInitializer : public cbm::algo::kf::SetupInitializer {
   public:
    /// \brief  Initializes the instance
    /// \param  mvd   Is MVD used
    /// \param  sts   Is STS used
    /// \param  much  Is MuCh used
    /// \param  trd   Is TRD used
    /// \param  tof   Is TOF used
    bool InitProperties() override;

    /// \brief  Enables/disables detector subsystems in the setup
    /// \param  mvd   Is MVD used
    /// \param  sts   Is STS used
    /// \param  much  Is MuCh used
    /// \param  trd   Is TRD used
    /// \param  tof   Is TOF used
    void Use(bool mvd, bool sts, bool much, bool trd, bool tof);

   private:
    // Material map creator properties (TODO: Provide setters, if needed)
    static constexpr double kMatCreatorPitch{0.1};    ///< Material budget map minimal bin size [cm]
    static constexpr int kMatCreatorMaxNbins{100};    ///< Max number of bins in the material budget map in x(y) axis
    static constexpr int kMatCreatorNrays{3};         ///< Number of rays per dimension for the material budget
    static constexpr bool kMatCreatorSafeMode{true};  ///< Safe mode of the material map creation
    static constexpr double kTargFieldInitStep{2.5};  ///< Step between nodes in the target field initialization [cm]
    static constexpr double kTargMaterialOffset{1};   ///< Offset between target upper limit and its material zMax [cm]

    bool fbUseMvd{false};   ///< Are MVD stations included in the tracking setup
    bool fbUseSts{false};   ///< Are STS stations included in the tracking setup
    bool fbUseMuch{false};  ///< Are MuCh stations included in the tracking setup
    bool fbUseTrd{false};   ///< Are TRD stations included in the tracking setup
    bool fbUseTof{false};   ///< Are TOF stations included in the tracking setup
  };
}  // namespace cbm::kf
